#include "ehs/io/audio/AudioDevice_ALSA.h"
#include "ehs/EHS.h"
#include "ehs/Log.h"
#include "ehs/system/Semaphore.h"

#include <spa/utils/list.h>
#include <spa/utils/result.h>
#include <spa/utils/dict.h>

#include "ehs/io/audio/AudioDevice.h"

namespace ehs
{
	Array<AudioDevice> AudioDevice::devices;

	void AudioDevice::RegistryEventGlobal(void *user_data, UInt_32 id, UInt_32 permissions, const char *type, UInt_32 version, const spa_dict *props)
	{
		const Str_8 mediaClass = spa_dict_lookup(props, PW_KEY_MEDIA_CLASS);

		if (mediaClass.Size() && (mediaClass == "Audio/Sink" || mediaClass == "Audio/Source"))
		{
			AudioDevice device;

			if (mediaClass == "Audio/Sink")
				device.type = AudioDeviceType::OUTPUT;
			else if (mediaClass == "Audio/Source")
				device.type = AudioDeviceType::INPUT;

			device.id = id;
			device.name = spa_dict_lookup(props, PW_KEY_NODE_NAME);

			EHS_LOG_INT(LogType::INFO, 1, "\nDevice Name: " + device.name + "\nId: " + Str_8::FromNum(device.id));

			const pw_metadata *metadata = pw_core_get_metadata();

			devices.Push(device);
		}
	}

	void AudioDevice::RegistryEventGlobalRemove(void *user_data, UInt_32 id)
	{
	}

	AudioDevice::~AudioDevice()
	{
	}

	AudioDevice::AudioDevice()
	{
	}

	AudioDevice::AudioDevice(AudioDevice&& device) noexcept
		: BaseAudioDevice(device)
	{
	}

	AudioDevice::AudioDevice(const AudioDevice& device)
		: BaseAudioDevice(device)
	{
	}

	AudioDevice& AudioDevice::operator=(AudioDevice&& device) noexcept
	{
		if (this == &device)
			return *this;

		BaseAudioDevice::operator=(device);

		return *this;
	}

	AudioDevice& AudioDevice::operator=(const AudioDevice& device)
	{
		if (this == &device)
			return *this;

		BaseAudioDevice::operator=(device);

		return *this;
	}

	void AudioDevice::Release()
	{
		if (!IsValid())
			return;
	}

	void AudioDevice::OpenStream()
	{
		if (streaming)
			return;

		pw_init(nullptr, nullptr);

		loop = pw_loop_new(nullptr);
		context = pw_context_new(loop, nullptr, 0);
		core = pw_context_connect(context, nullptr, 0);

		spa_audio_info_raw info = {
			.format = SPA_AUDIO_FORMAT_F32,
			.rate = 48000,
			.channels = 2,
			.position = {
				SPA_AUDIO_CHANNEL_FL,
				SPA_AUDIO_CHANNEL_FR,
				SPA_AUDIO_CHANNEL_FC,
				SPA_AUDIO_CHANNEL_LFE,
				SPA_AUDIO_CHANNEL_SL,
				SPA_AUDIO_CHANNEL_SR,
				SPA_AUDIO_CHANNEL_RL,
				SPA_AUDIO_CHANNEL_RR
			}
		};

		pw_properties *props = pw_properties_new(
			PW_KEY_MEDIA_TYPE, "Audio",
			PW_KEY_MEDIA_CATEGORY, "Playback",
			PW_KEY_MEDIA_ROLE, "Game",
			nullptr
		);

		spa_pod_builder b = {};
		Byte buffer[1024];

		spa_pod_builder_init(&b, buffer, sizeof(buffer));

		const spa_pod *pod[] = {spa_format_audio_raw_build(&b, SPA_PARAM_EnumFormat, &info)};

		pw_stream_connect(
			stream,
			PW_DIRECTION_OUTPUT,
			PW_ID_ANY,
			(pw_stream_flags)(PW_STREAM_FLAG_AUTOCONNECT | PW_STREAM_FLAG_MAP_BUFFERS),
			pod,
			1
		);

		streaming = true;
	}

	void AudioDevice::CloseStream()
	{
		if (!streaming)
			return;

		pw_stream_destroy(stream);
		pw_core_disconnect(core);
		pw_context_destroy(context);
		pw_loop_destroy(loop);
		pw_deinit();

		streaming = false;
	}

	UInt_64 AudioDevice::SendStream(void *data, const UInt_64 size)
	{
		while (pw_loop_iterate(loop, 0));

		if (pw_stream_get_state(stream, nullptr) != PW_STREAM_STATE_STREAMING)
			return 0;

		pw_buffer *buf = pw_stream_dequeue_buffer(stream);
		if (!buf)
			return 0;

		UInt_64 out = 0;

		spa_data &result = buf->buffer->datas[0];

		if (size > result.maxsize)
			out = result.maxsize;
		else
			out = size;

		Util::Copy(result.data, data, out);

		pw_stream_queue_buffer(stream, buf);

		return out;
	}

	bool AudioDevice::IsValid() const
	{
		return true;
	}

	AudioDevice AudioDevice::GetDefault(const AudioDeviceType type)
	{
		pw_init(nullptr, nullptr);

		pw_loop *loop = pw_loop_new(nullptr);
		pw_context *context = pw_context_new(loop, nullptr, 0);
		pw_core *core = pw_context_connect(context, nullptr, 0);
		pw_registry *registry = pw_core_get_registry(core, PW_VERSION_REGISTRY, 0);
		spa_hook listener = {};

		constexpr pw_registry_events events = {
			.version = PW_VERSION_REGISTRY_EVENTS,
			.global = RegistryEventGlobal,
			.global_remove = RegistryEventGlobalRemove
		};

		pw_registry_add_listener(registry, &listener, &events, nullptr);

		while (pw_loop_iterate(loop, 10));

		constexpr pw_metadata_events

		return {};
	}

	Array<AudioDevice> AudioDevice::Get(const AudioDeviceType type, const AudioDeviceState state)
	{
		return {};
	}
}
